package msgpack_test

import (
	"bytes"
	"encoding/json"
	"fmt"
	"os"
	"path/filepath"
	"strconv"
	"strings"
	"testing"

	"emd-api/internal/msgpack"

	"carbonblack.com/protocols/psc/ems/server/emdstatus"
	"github.com/rs/zerolog/log"
)

const (
	base10   = 10
	path     = "../../data/dev_status_req_msgpack.bin"
	nilValue = "NIL"
)

var (
	data   []byte                     //nolint:gochecknoglobals // for testing only
	parser *msgpack.MessagePackParser //nolint:gochecknoglobals // for testing only

	//nolint:gochecknoglobals // for testing only
	inResults = map[string]interface{}{
		msgpack.FieldNameTimeSent:                    uint64(1000),
		msgpack.FieldNameRequestVersion:              int8(19),
		msgpack.FieldNameDeviceGUID:                  "498d862e93c856dfd6e72c1af0ab8a438c66fc91b4b22f6dbd60d18f1f3ef85f", //nolint:lll // pragma: allowlist secret
		msgpack.FieldNameRegistrationID:              "testOrgId",
		msgpack.FieldNameDeviceType:                  int8(1),
		msgpack.FieldNameMessageType:                 uint16(8193),
		msgpack.FieldNameSubCode:                     int8(0),
		msgpack.FieldNameMessageID:                   "testMessageId",
		msgpack.FieldNameAgentLocation:               int8(1),
		msgpack.FieldNameLocalAddress:                "1234",
		msgpack.FieldNameMacAddress:                  "ffff:ffff:ffff:ffff",
		msgpack.FieldNameProtocolFamily:              int8(2),
		msgpack.FieldNameEchoDummyResponse:           false,
		msgpack.FieldNameLastUpdateTime:              uint64(1223372036854775807),
		msgpack.FieldNameLastLiveQueryTime:           uint64(1223372036854775807),
		msgpack.FieldNamePlatformType:                int8(2),
		msgpack.FieldNameOsMajorVersion:              int8(5),
		msgpack.FieldNameOsMinorVersion:              int8(1),
		msgpack.FieldNameMaintenanceRelease:          int8(111),
		msgpack.FieldNameHostname:                    "testHost",
		msgpack.FieldNameProductVersion:              "clientv55",
		msgpack.FieldNameLocale:                      "testLocale",
		msgpack.FieldNamePolicyBlockedHashes:         uint32(9999999),
		msgpack.FieldNamePolicyRestrictedHashes:      uint32(9999999),
		msgpack.FieldNameRunBlockedHashes:            uint32(9999999),
		msgpack.FieldNameRunRestrictedHashes:         uint32(9999999),
		msgpack.FieldNameNewPolicyBlockedHashes:      uint32(9999999),
		msgpack.FieldNameNewPolicyRestrictedHashes:   uint32(9999999),
		msgpack.FieldNameNewRunBlockedHashes:         uint32(9999999),
		msgpack.FieldNameNewRunRestrictedHashes:      uint32(9999999),
		msgpack.FieldNamePreviousReportTime:          uint16(10000),
		msgpack.FieldNameRootkitDetected:             true,
		msgpack.FieldNameQuarantined:                 false,
		msgpack.FieldNameEmail:                       "testEmail",
		msgpack.FieldNameSensorStatus:                []interface{}{int8(0), int8(102)},
		msgpack.FieldNameLastSensorResetTime:         uint16(10000),
		msgpack.FieldNameAvVersion:                   "1.2.3.4",
		msgpack.FieldNameAvVersionMinor:              "5",
		msgpack.FieldNameAvStatus:                    []interface{}{int8(20), int8(1)},
		msgpack.FieldNameAvLastScanTime:              uint16(10000),
		msgpack.FieldNameAvUpdateServers:             []interface{}{"10.10.10.10:7000", "11.11.11.11:7000"},
		msgpack.FieldNameAvReachableServers:          []interface{}{"10.10.10.10:7000", "11.11.11.11:7000"},
		msgpack.FieldNamePolicyName:                  "testPolicy",
		msgpack.FieldNamePolicyTimeStamp:             uint16(10000),
		msgpack.FieldNameExternallyAssignedDeviceID:  "testId",
		msgpack.FieldNameVdiProvider:                 "testProvider",
		msgpack.FieldNameInfrastructureProvider:      "testProvider",
		msgpack.FieldNameVirtualizationProvider:      "testProvider",
		msgpack.FieldNameIsVirtualMachine:            false,
		msgpack.FieldNameOsBuildNumber:               int8(10),
		msgpack.FieldNameLoginUserName:               "testUser",
		msgpack.FieldNameUnifiedEndpointManagementID: "testId",
	}
)

func TestMain(m *testing.M) {
	setup()

	code := m.Run()

	teardown()
	os.Exit(code)
}

func setup() {
	parser = msgpack.NewMessagePackParser()
	data, _ = os.ReadFile(filepath.Clean(path))
}

func teardown() {}

// setFieldWithSprint is an internal/msgpack/parser#setField implementation that uses a fmt.Sprint to produce a string.
func setFieldWithSprint(results map[string]string, field string, value interface{}) {
	stringValue := fmt.Sprint(value)
	if stringValue != nilValue {
		results[field] = stringValue
	}
}

// setFieldWithJSONMarshal is an internal/msgpack/parser#setField implementation that uses a json.Marshal to produce a
// string.
func setFieldWithJSONMarshal(results map[string]string, field string, value interface{}) error {
	bytes, err := json.Marshal(value)
	if err != nil {
		return fmt.Errorf("error parsing field %s type %T value %v: %w", field, value, value, err)
	}

	stringValue := string(bytes)
	if stringValue != fmt.Sprintf("\"%s\"", nilValue) {
		results[field] = stringValue
	}

	return nil
}

func setArrayField(typedValue []interface{}, field string) string {
	stringBuilder := make([]string, 0, len(typedValue))

	for _, arrayValue := range typedValue {
		switch typedValue := arrayValue.(type) {
		case string:
			stringBuilder = append(stringBuilder, "\""+typedValue+"\"")
		case int8:
			stringBuilder = append(stringBuilder, strconv.Itoa(int(typedValue)))
		default:
			log.Warn().Msgf("setting unexpected type %T for field %s", typedValue, field)
			stringBuilder = append(stringBuilder, fmt.Sprint(typedValue))
		}
	}

	return "[" + strings.Join(stringBuilder, ",") + "]"
}

// setFieldWithTypeSwitch is an internal/msgpack/parser#setField implementation that uses a type switch to produce a
// string.
func setFieldWithTypeSwitch(results map[string]string, field string, value interface{}) {
	switch typedValue := value.(type) {
	case []interface{}:
		results[field] = setArrayField(typedValue, field)
	case bool:
		results[field] = strconv.FormatBool(typedValue)
	case int8:
		results[field] = strconv.Itoa(int(typedValue))
	case string:
		if typedValue != nilValue {
			results[field] = "\"" + typedValue + "\""
		}
	case uint16:
		results[field] = strconv.FormatUint(uint64(typedValue), base10)
	case uint32:
		results[field] = strconv.FormatUint(uint64(typedValue), base10)
	case uint64:
		results[field] = strconv.FormatUint(typedValue, base10)
	}
}

// setFieldWithTypeSwitchAndSomeJSON is an internal/msgpack/parser#setField implementation that uses a type switch to
// produce a string, falling back to json marshalling for arrays.
func setFieldWithTypeSwitchAndSomeJSON(results map[string]string, field string, value interface{}) error {
	switch typedValue := value.(type) {
	case []interface{}:
		bytes, err := json.Marshal(typedValue)
		if err != nil {
			return fmt.Errorf("error parsing field %s type %T value %v: %w", field, value, value, err)
		}

		results[field] = string(bytes)
	case bool:
		results[field] = strconv.FormatBool(typedValue)
	case int8:
		results[field] = strconv.Itoa(int(typedValue))
	case string:
		if typedValue != nilValue {
			results[field] = "\"" + typedValue + "\""
		}
	case uint16:
		results[field] = strconv.FormatUint(uint64(typedValue), base10)
	case uint32:
		results[field] = strconv.FormatUint(uint64(typedValue), base10)
	case uint64:
		results[field] = strconv.FormatUint(typedValue, base10)
	}

	return nil
}

// setFieldWithTypeSwitchAndSomeSprint is an internal/msgpack/parser#setField implementation that uses a type switch to
// produce a string, falling back to fmt.Sprint for arrays.
func setFieldWithTypeSwitchAndSomeSprint(results map[string]string, field string, value interface{}) {
	switch typedValue := value.(type) {
	case []interface{}:
		results[field] = fmt.Sprint(typedValue)
	case bool:
		results[field] = strconv.FormatBool(typedValue)
	case int8:
		results[field] = strconv.Itoa(int(typedValue))
	case string:
		if typedValue != nilValue {
			results[field] = "\"" + typedValue + "\""
		}
	case uint16:
		results[field] = strconv.FormatUint(uint64(typedValue), base10)
	case uint32:
		results[field] = strconv.FormatUint(uint64(typedValue), base10)
	case uint64:
		results[field] = strconv.FormatUint(typedValue, base10)
	}
}

// ===== BENCHMARKS ===== //

//nolint:thelper // disabled benchmark
func XBenchmarkSprint(b *testing.B) {
	b.SetBytes(512)

	for n := 0; n < b.N; n++ {
		outResults := make(map[string]string, len(inResults))

		for field, value := range inResults {
			setFieldWithSprint(outResults, field, value)
		}
	}
}

//nolint:thelper // disabled benchmark
func XBenchmarkJSONMarshal(b *testing.B) {
	b.SetBytes(512)

	for n := 0; n < b.N; n++ {
		outResults := make(map[string]string, len(inResults))

		for field, value := range inResults {
			if err := setFieldWithJSONMarshal(outResults, field, value); err != nil {
				b.Error(err)
			}
		}
	}
}

//nolint:thelper // disabled benchmark
func XBenchmarkTypeSwitch(b *testing.B) {
	b.SetBytes(512)

	for n := 0; n < b.N; n++ {
		outResults := make(map[string]string, len(inResults))

		for field, value := range inResults {
			setFieldWithTypeSwitch(outResults, field, value)
		}
	}
}

//nolint:thelper // disabled benchmark
func XBenchmarkTypeSwitchWithSomeJSON(b *testing.B) {
	b.SetBytes(512)

	for n := 0; n < b.N; n++ {
		outResults := make(map[string]string, len(inResults))

		for field, value := range inResults {
			if err := setFieldWithTypeSwitchAndSomeJSON(outResults, field, value); err != nil {
				b.Error(err)
			}
		}
	}
}

//nolint:thelper // disabled benchmark
func XBenchmarkTypeSwitchWithSomeSprint(b *testing.B) {
	b.SetBytes(512)

	for n := 0; n < b.N; n++ {
		outResults := make(map[string]string, len(inResults))

		for field, value := range inResults {
			setFieldWithTypeSwitchAndSomeSprint(outResults, field, value)
		}
	}
}

//nolint:thelper // disabled benchmark
func XBenchmarkParser(b *testing.B) {
	b.SetBytes(int64(len(data)))

	resource := &emdstatus.DeviceStatusMsg{Metadata: map[string]string{}}

	for n := 0; n < b.N; n++ {
		if err := parser.ParseStatusMessage(resource, bytes.NewReader(data), &log.Logger); err != nil {
			b.Error(err)
		}
	}
}
