#!/bin/bash

# change to home dir for file create perms
cd ~ || exit 1

# generate random 32 byte hex guid
guid=$(xxd -l 32 -p /dev/urandom | tr -d ' \n'); echo "generated guid: ${guid}"

# create new org, dev, and device_group for testing
orgid=$(psql -c "INSERT INTO organization (org_type_id, name, create_time) VALUES (1, 'emds-e2e-test-org', NOW()) RETURNING org_id;" | grep -E '^ +[0-9]' | awk '{print $1}'); echo "generated org: ${orgid}"
devid=$(psql -c "INSERT INTO device (device_type_id, device_guid, properties) VALUES (1, '${guid}', '{\"Maintenance Release\":\"1\"}') RETURNING device_id;" | grep -E '^ +[0-9]' | awk '{print $1}'); echo "generated device: ${devid}"
dgid=$(psql -c "INSERT INTO device_group (org_id, policy_control, is_default, name, priority_level) VALUES (${orgid}, TRUE, TRUE, 'default', 1) RETURNING device_group_id;" | grep -E '^ +[0-9]' | awk '{print $1}'); echo "generated group: ${dgid}"
psql -c "INSERT INTO device_group_device (device_group_id, device_id) VALUES (${dgid}, ${devid})"

env="${ENV}"
regid="${orgid}-${devid}"
mpregid=$(echo '"'"${regid}"'"' | json2msgpack | xxd -p)
mpstatus1=a800000000000003e813da0020
mpstatus2=01cd200100c001c0c2c080020501
mpstatus3=a874657374486f7374a9636c69656e74763535c0000000000000000000c2c2c092006600c0c0c000c0c0c0c081b049735669727475616c4d616368696e65c200c0c0

# get org-key to set on org so emd msg is properly augmented
orgkey=$(curl \
  -H "Registration-ID: ${regid}" \
  -H "Device-GUID: ${guid}" \
  -H "Device-Environment: ${env}" \
  -H "Accept: application/x-protobuf" \
  -sS -D - \
  ems-device-auth/ems/v1/device_auth \
  -o /dev/null | \
    grep X-Orgkey | \
    sed 's/X-Orgkey: //' | \
    xargs); echo "got org-key: ${orgkey}"
# set org-key on org
psql -c "UPDATE organization SET org_key='${orgkey}' WHERE org_id=${orgid};"
# evict org
curl -v -X DELETE "ems-device-auth/ems/v1/orgs/${orgkey}/org/${orgid}"

# get device-context to pass to emd-api status api
ctx=$(curl \
  -H "Registration-ID: ${regid}" \
  -H "Device-GUID: ${guid}" \
  -H "Device-Environment: ${env}" \
  -H "Accept: application/x-protobuf" \
  -sS -D - \
  ems-device-auth/ems/v1/device_auth \
  -o /dev/null | \
    grep Device-Context | \
    sed 's/Device-Context: //' | \
    xargs); echo "got device-context: ${ctx}"

# get current maintenance release version for device
version=$(curl \
  -H "Registration-ID: ${regid}" \
  -H "Device-GUID: ${guid}" \
  -H "Device-Environment: ${env}" \
  -sS -D - \
  ems-device-auth/ems/v1/device_auth \
  -o /dev/null | \
    grep Device-Context | \
    sed 's/Device-Context: //' | \
    jq '.device.properties."Maintenance Release"'); echo "got version: ${version}"

# increment maintenance release version, mod 10 to keep it under 10, pad with 0 for msgpack
next="0$(( (${version//\"/} + 1) % 10 ))"

# generate msgpack status message
mpstatusmsg=${mpstatus1}${guid}${mpregid}${mpstatus2}${next}${mpstatus3}
echo "${mpstatusmsg}" | xxd -r -p > emdstatusmsg

# post to emd-api status
curl -v -H "Device-Context: $ctx" --form message='@emdstatusmsg' --output - emd-api:8080/emd-api/v1/status; echo "sent status msg: ${mpstatusmsg}"

# confirm change to device via device-auth, try 3 times
echo "confirming status update propagated through emds, looking for ${next} version"
try=1; while [ "${try}" -le 3 ]; do
  check=$(curl \
    -H "Registration-ID: ${regid}" \
    -H "Device-GUID: ${guid}" \
    -H "Device-Environment: ${env}" \
    -sS -D - \
    ems-device-auth/ems/v1/device_auth \
    -o /dev/null | \
      grep Device-Context | \
      sed 's/Device-Context: //' | \
      jq '.device.properties."Maintenance Release"'); echo "got version: ${check}"
  [ "0${check//\"/}" = "${next}" ] && echo "success!" && break
  try=$(( "${try}" + 1 ))
  echo "trying again in 1 second"; sleep 1
done

# cleanup db
echo "cleaning up db"
psql -c "DELETE FROM device_group_device WHERE device_group_id=${dgid};"; echo "deleted device_group_device joins"
psql -c "DELETE FROM device WHERE device_id=${devid};"; echo "deleted device"
psql -c "DELETE FROM device_group WHERE org_id=${orgid};"; echo "deleted device_group"
psql -c "DELETE FROM organization WHERE org_id=${orgid};"; echo "deleted org"
echo "done cleaning"

# do one last check to ensure script exits with proper status
[ "0${check//\"/}" = $next ]
