//go:build integration
// +build integration

package integration

import (
	"bytes"
	"encoding/base64"
	"io"
	"mime/multipart"
	"net/http"
	"os"
	"path/filepath"
	"testing"

	"carbonblack.com/protocols/psc/ems/server/deviceauth"
	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
	"google.golang.org/protobuf/proto"
)

var host = "http://emd-api:8080/emd-api"

func TestEmdApi(t *testing.T) {
	t.Parallel()
	RegisterFailHandler(Fail)
	RunSpecs(t, "emd-api")
}

var _ = Describe("integration test", func() {
	parseMsgpackFile := func(fp string) io.Reader {
		// Test data -> https://gitlab.bit9.local/psc/platform/ems/emd/api/-/snippets/384.
		data, err := os.ReadFile(filepath.Clean(fp))
		Expect(err).To(BeNil())
		return bytes.NewReader(data)
	}

	Context("emd-api APIs", func() {
		It("should report healthy on healthCheck request", func() {
			req, err := http.NewRequest(http.MethodGet, host+"/healthCheck", nil)
			Expect(err).To(BeNil())
			client := http.Client{}
			res, err := client.Do(req)
			Expect(err).To(BeNil())
			Expect(res.StatusCode).To(Equal(http.StatusNoContent))
		})

		It("should respond accepted for status request with all fields", func() {
			data := parseMsgpackFile("../data/dev_status_req_msgpack.bin")
			res, err := sendMsgpackReq(host+"/v1/status", data)
			Expect(err).To(BeNil())
			Expect(res).ToNot(BeNil())
			Expect(res.StatusCode).To(Equal(http.StatusAccepted))
		})

		It("should respond accepted for status request with only mandatory fields", func() {
			data := parseMsgpackFile("../data/dev_status_req_only_mandatory_fields.bin")
			res, err := sendMsgpackReq(host+"/v1/status", data)
			Expect(err).To(BeNil())
			Expect(res).ToNot(BeNil())
			Expect(res.StatusCode).To(Equal(http.StatusAccepted))
		})

		It("should respond accepted for status request with later version fields", func() {
			data := parseMsgpackFile("../data/dev_status_req_v18.bin")
			res, err := sendMsgpackReq(host+"/v1/status", data)
			Expect(err).To(BeNil())
			Expect(res).ToNot(BeNil())
			Expect(res.StatusCode).To(Equal(http.StatusAccepted))
		})

		It("should respond with server error for invalid status request", func() {
			data := parseMsgpackFile("../data/invalid_dev_status_req.bin")
			res, err := sendMsgpackReq(host+"/v1/status", data)
			defer res.Body.Close()

			Expect(err).To(BeNil())
			Expect(res).ToNot(BeNil())
			Expect(res.StatusCode).To(Equal(http.StatusInternalServerError))
		})
	})
})

func sendMsgpackReq(url string, data io.Reader) (res *http.Response, err error) {
	// Prepare a form that you will submit to that URL.
	var b bytes.Buffer
	var fw io.Writer
	w := multipart.NewWriter(&b)
	if fw, err = w.CreateFormField("message"); err != nil {
		return
	}
	if _, err = io.Copy(fw, data); err != nil {
		return
	}
	// close the multipart writer for adding the terminating boundary
	w.Close()

	req, err := http.NewRequest(http.MethodPost, url, &b)
	if err != nil {
		return
	}
	// Don't forget to set the content type, this will contain the boundary
	req.Header.Set("Content-Type", w.FormDataContentType())
	devCtx, err := proto.Marshal(&deviceauth.Context{Device: getProtoDevice(), DeviceGroup: getProtoDeviceGroup(), Organization: getProtoOrg()})
	req.Header.Set("Device-Context", base64.StdEncoding.EncodeToString(devCtx))

	// Submit the request
	client := http.Client{}
	res, err = client.Do(req)

	return
}

func getProtoOrg() *deviceauth.Organization {
	return &deviceauth.Organization{
		Id:                             1,
		OrgType:                        0,
		Name:                           "test@test.com",
		AdminContactId:                 0,
		BillingContactId:               0,
		RegistrationPassword:           "",
		AnalyticsProfile:               "",
		AnalyticsVersion:               "",
		StorageProfile:                 "",
		CreateTime:                     nil,
		UpdateVersion:                  0,
		Mode:                           0,
		TrialEndTime:                   nil,
		CreateByAdmin:                  "",
		PaidStartTime:                  nil,
		ChangeTimestamps:               nil,
		KillSwitchOn:                   false,
		RansomwareConfig:               "",
		UninstallCode:                  "",
		LicenseEndpoints:               0,
		LicenseExpires:                 nil,
		OrgKey:                         "test123",
		EffectiveThEndpoints:           0,
		DeploymentTypeChangeTimestamps: nil,
	}
}

func getProtoDeviceGroup() *deviceauth.DeviceGroup {
	return &deviceauth.DeviceGroup{
		Id:               1,
		OrgId:            1,
		PolicyControl:    false,
		Default:          false,
		Priority:         0,
		Name:             "",
		Description:      "",
		AnalyticsProfile: "",
		AnalyticsVersion: "",
		AdminContactId:   0,
		Policy:           "",
		PolicyV2:         nil,
		UpdateVersion:    0,
		PriorityLevel:    0,
		ChangeTimestamps: nil,
	}
}

func getProtoDevice() *deviceauth.Device {
	return &deviceauth.Device{
		Id:               1,
		DeviceType:       0,
		DeviceOwnerId:    0,
		DeviceGuid:       "test123",
		Name:             "",
		Email:            "",
		EmailType:        0,
		DeviceSessionId:  "",
		AssignedToId:     0,
		AssignedToName:   "",
		CreateTime:       nil,
		UpdateVersion:    0,
		TimeCorrectionMs: 0,
		ReplicaDeviceId:  0,
		ChangeTimestamps: nil,
		KillSwitchOn:     false,
		MsmGroupId:       0,
		MsmGroupName:     "",
		Properties:       nil,
		OrgKey:           "test123",
		DeploymentType:   0,
	}
}
